/*
 *	Copyright (C) Jan Bobrowski
 *
 *	This library is free software; you can redistribute and modify
 *	it under the terms of the GNU Lesser General Public License
 *	version 2.1 as published by the Free Software Foundation.
 */

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <poll.h>
#include <sys/ioctl.h>
#include <signal.h>
#include "kbd.h"

#define elemof(T) (sizeof T/sizeof*T)

struct square {
	int x, y;
	char chr;
	char color;
	int p1, p2;
	struct square *nx, **ppv;
};

struct square tbl[] = {
	{60,5,'a',1},
	{10,15,'a',2},
	{20,10,'a',3},
	{70,12,'a',4},
};

int w = 80, h = 24, sqw = 8, sqh = 4;

#define ESC "\33"
#define CSI "\33["

static void disp()
{
	int y;
	printf("\16");
	for(y = 0; y < h; y++) {
		int n;
		struct square *lst = 0;
		for(n = 0; n < elemof(tbl); n++) {
			struct square **p, *s = &tbl[n], *q;
			int r1, r2;
			if(y < s->y || y >= s->y + sqh)
				continue;
			r1 = 0;
			for(p = &lst; (q = *p); p = &q->nx) {
				if(s->x <= q->x)
					break;
				r1 = q->x + sqw;
			}
			if(q && s->x == q->x)
				continue;
			s->p1 = r1<s->x ? s->x : r1;
			s->p2 = s->x + sqw;
			r2 = q ? q->x : w;
			if(r2 < s->p2) {
				s->p2 = r2;
				if(s->p2 <= s->p1)
					continue;
			}
			s->nx = q;
			*p = s;
		}
		printf(CSI "%dH", y + 1);
		{
			struct square *p = lst;
			int r = 0;
			while(p) {
//				if(p->p1 > r) printf(CSI "%d@", p->p1 - r);
				for(n = r; n < p->p1; n++)
					putchar(' ');
				printf(CSI "1;3%d;4%dm", p->color, p->color);
				for(n = p->p1; n < p->p2; n++)
					putchar(p->chr);
				r = p->p2; p = p->nx;
				printf(CSI "m");
			}
		if(r < w)
			printf(CSI "K");
		}
	}
	printf("\17");
}

void mouse()
{
	static int o = -1;
	static int ox, oy;
	const struct mouse_event *e = getmouse();
	switch(e->code) {
	case kmPRESS: {
		int n;
#define INSQR(m, s, n) (m >= s && m < s + n)
		for(n = 0; n < elemof(tbl); n++)
			if(INSQR(e->x, tbl[n].x, sqw) && INSQR(e->y, tbl[n].y, sqh)) {
				o=n;
				break;
			}
	} break;
	case kmWHEEL:
		sqh += e->wy;
		if(sqh < 2)
			sqh = 2;
		else if(sqh > 8)
			sqh = 8;
		sqw = 2*sqh;
	case kmRELEASE:
	case kmDRAG:
		if(o < 0)
			break;
		tbl[o].x += e->x - ox;
		tbl[o].y += e->y - oy;
		if(e->code == kmRELEASE)
			o = -1;
		break;
	}
	ox = e->x;
	oy = e->y;
}

void term_raw();
void term_rest();

void restore() {
	printf(CSI "?25h" CSI "?7r" CSI "?1002l" CSI "?1000l" ESC "8");
	term_rest();
}


volatile int rq_winch;

void getwinsz()
{
	struct winsize ws;
	ws.ws_col = 80;
	ws.ws_row = 24;
	rq_winch = 0;
	ioctl(1, TIOCGWINSZ, &ws);
	w = ws.ws_col;
	h = ws.ws_row;
}

void sigwinch(int n)
{
	rq_winch = 1;
}

int main()
{
	struct pollfd fd;
	fd.fd = 0;
	fd.events = POLLIN;

	getwinsz();
	disp();
	atexit(restore);
	term_raw();
	signal(SIGWINCH, sigwinch);
	setvbuf(stdout, 0, _IOFBF, BUFSIZ);
	printf(ESC "7" CSI "?1000h" CSI "?1002h" CSI "?7s" CSI "?7l" CSI "?25l");

	for(;;) {
		int k;
		fflush(stdout);
		poll(&fd, 1, -1);
		if(rq_winch) {
			getwinsz();
			disp();
		}
		if(!fd.revents)
			continue;
		for(;;) {
			k = getkey();
			if(k == kNOKEY)
				break;
			if(k == kERROR)
				return 1;
			if(k == kMORE) {
				usleep(KBD_WAIT_TO_MS * 1000);
				continue;
			}
			if(k == kESC || k == 'q' || k == 'Q')
				return 0;
			if(k == kMOUSE)
				mouse();
		}
		disp();
	}
}
